#include "group.h"

/**
 *
 * Beryl group plugin
 *
 * group.c
 *
 * Copyright : (C) 2006 by Patrick Niklaus, Roi Cohen, Danny Baumann
 * Authors: Patrick Niklaus <patrick.niklaus@googlemail.com>
 *          Roi Cohen       <roico@beryl-project.org>
 *          Danny Baumann   <maniac@beryl-project.org>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 **/

/* helper for activating a window after a certain time a slot has dragged over it */

static Bool
groupDragHoverTimeout(void* closure)
{
	CompWindow *w = (CompWindow *) closure;

	if (!w)
		return FALSE;

	GROUP_SCREEN(w->screen);

	activateWindow(w);
	gs->dragHoverTimeoutHandle = 0;

	return FALSE;
}

/*
 * groupFindWindowIndex
 *
 */
int
groupFindWindowIndex(CompWindow *w, GroupSelection *g)
{
	int i;

	for(i = 0; i < g->nWins; i++)
	{
		if(g->windows[i]->id == w->id)
			return i;
	}

	return -1;
}

/*
 * groupSyncWindows
 *
 */
void groupSyncWindows(GroupSelection *group)
{
       int i;
       for (i = 0; i < group->nWins; i++) {
               CompWindow *w = group->windows[i];
               syncWindowPosition(w);
       }
}

/*
 * groupRaiseWindows
 *
 */
void
groupRaiseWindows(CompWindow * top, GroupSelection *group)
{
	int i;
	for (i = 0; i < group->nWins; i++) {
		CompWindow *w = group->windows[i];
		if (w->id == top->id)
			continue;
		restackWindowBelow(w, top);
	}
}

/*
 * groupDeleteGroupWindow
 *
 */
void groupDeleteGroupWindow(CompDisplay * d, CompWindow * w)
{
	GROUP_WINDOW(w);
	GROUP_SCREEN(w->screen);
	GroupSelection *group = gw->group;

	if (!group)
		return;

	if (group->tabBar) {
		if (gw->slot)
		{
			if( gs->draggedSlot && gs->dragged && gs->draggedSlot->window->id == w->id)
				groupUnhookTabBarSlot(group->tabBar, gw->slot, FALSE);
			else
		        	groupDeleteTabBarSlot(group->tabBar, gw->slot);
		}
	}

	if (group->nWins && group->windows) {
		CompWindow **buf = group->windows;

		group->windows = (CompWindow **) calloc(group->nWins - 1, sizeof(CompWindow *));

		int counter = 0;
		int i;
		for (i = 0; i < group->nWins; i++) {
			if (buf[i]->id == w->id)
				continue;
			group->windows[counter++] = buf[i];
		}
		group->nWins = counter;

		if (group->nWins == 1 &&
		    gs->opt[GROUP_SCREEN_OPTION_AUTO_UNGROUP].value.b) {
			if (group->changeTab) {
				/* a change animation is pending: this most
				   likely means that a window must be moved
				   back onscreen, so we do that here */

				groupMoveTabOffscreen(group->windows[0], group, FALSE);
			}
			groupDeleteGroup(d, group);
		} else if (group->nWins <= 0) {
			free(group->windows);
			group->windows = NULL;
			groupDeleteGroup(d, group);
		}

		free(buf);

		damageWindowOutputExtents(w);
		gw->group = NULL;
		updateWindowOutputExtents(w);
	}
}

/*
 * groupDeleteGroup
 *
 */
void groupDeleteGroup(CompDisplay * d, GroupSelection *group)
{
	GROUP_DISPLAY(d);

	if (group->windows != NULL) {
		if (group->tabBar) {
			groupUntabGroup(group);
			group->ungroup = TRUE;
			return;
		}

		int i;
		for (i = 0; i < group->nWins; i++) {
			CompWindow *cw = group->windows[i];
			GROUP_WINDOW(cw);

			damageWindowOutputExtents(cw);
			gw->group = NULL;
			updateWindowOutputExtents(cw);
		}
		free(group->windows);
		group->windows = NULL;
	}

	GroupSelection *prev = group->prev;
	GroupSelection *next = group->next;

	// relink stack
	if (prev || next) {
		if (prev) {
			if (next)
				prev->next = next;
			else {
				prev->next = NULL;
				//gd->revGroups = prev;
			}
		}
		if (next) {
			if (prev)
				next->prev = prev;
			else {
				next->prev = NULL;
				gd->groups = next;
			}
		}
	} else {
		gd->groups = NULL;
	}
	free(group);
}

/*
 * groupAddWindowToGroup
 *
 */
void
groupAddWindowToGroup(CompDisplay * d, CompWindow * w, GroupSelection *group)
{
	GROUP_DISPLAY(d);
	GROUP_WINDOW(w);

	if(group && (gw->group == group))
		return;

	if (gw->group)
		groupDeleteGroupWindow(d, w);

	if (group) {
		group->windows = (CompWindow **) realloc(group->windows, sizeof(CompWindow *) * (group->nWins + 1));
		group->windows[group->nWins] = w;
		group->nWins++;

		gw->group = group;
		updateWindowOutputExtents(w);

		if (group->tabBar && group->topTab)
		{
			CompWindow *topTab = TOP_TAB(group);
			
			if(!gw->slot)
				groupCreateSlot(group, w);
		
			gw->destination.x = WIN_X(topTab) + (WIN_WIDTH(topTab) / 2) - (WIN_WIDTH(w) / 2);
			gw->destination.y = WIN_Y(topTab) + (WIN_HEIGHT(topTab) / 2) - (WIN_HEIGHT(w) / 2);
			gw->mainTabOffset.x = WIN_X(w) - gw->destination.x;
			gw->mainTabOffset.y = WIN_Y(w) - gw->destination.y;
			gw->orgPos.x = WIN_X(w);
			gw->orgPos.y = WIN_Y(w);

			gw->tx = gw->ty = gw->xVelocity = gw->yVelocity = 0.0f;
		
			gw->animateState = IS_ANIMATED;

			groupStartTabbingAnimation(group, TRUE);
			
			damageScreen(w->screen);
		}
	} else {
		GroupSelection *g = malloc(sizeof(GroupSelection));

		g->windows = (CompWindow **) calloc(1, sizeof(CompWindow *));
		g->windows[0] = w;
		g->nWins = 1;
		g->topTab = NULL;
		g->prevTopTab = NULL;
		g->nextTopTab = NULL;
		g->doTabbing = FALSE;
		g->changeAnimationTime = 0;
		g->changeAnimationDirection = 0;
		g->changeState = PaintOff;
		g->tabbingState = PaintOff;
		g->ungroup = FALSE;
		g->changeTab = FALSE;
		g->tabBar = NULL;

		g->grabWindow = None;
		g->grabMask = 0;
		
		g->inputPrevention.id = None;
		g->inputPrevention.mapped = FALSE;

		g->oldTopTabCenterX = 0;
		g->oldTopTabCenterY = 0;

		// glow color
		srand(time(NULL));
		g->color[0] = rand() % 0xFFFF;
		g->color[1] = rand() % 0xFFFF;
		g->color[2] = rand() % 0xFFFF;
		g->color[3] = 0xFFFF;

		// relink stack
		if (gd->groups) {
			gd->groups->prev = g;
			g->next = gd->groups;
			g->prev = NULL;
			gd->groups = g;
		} else {
			g->prev = NULL;
			g->next = NULL;
			gd->groups = g;
		}

		gw->group = g;
		updateWindowOutputExtents(w);
	}
}

/*
 * groupGroupWindows
 *
 */
Bool
groupGroupWindows(CompDisplay * d, CompAction * action,
		  CompActionState state, CompOption * option, int nOption)
{
	GROUP_DISPLAY(d);

	if (gd->tmpSel.nWins > 0) {
		int i;
		CompWindow *cw;
		GroupSelection *group = NULL;

		for(i = 0; i < gd->tmpSel.nWins; i++)
		{
			cw = gd->tmpSel.windows[i];
			GROUP_WINDOW(cw);

			if (gw->group)
			{
				if (!group)
					group = gw->group;

				if (group != gw->group) {
					group = NULL;
					break;
				}
			}
		}

		// we need to do one first to get the pointer of a new group
		cw = gd->tmpSel.windows[0];
		groupAddWindowToGroup(d, cw, group);

		GROUP_WINDOW (cw);
		gw->inSelection = FALSE;
		damageScreen(cw->screen);
		group = gw->group;

		for (i = 1; i < gd->tmpSel.nWins; i++) {
			cw = gd->tmpSel.windows[i];
			GROUP_WINDOW(cw);

			groupAddWindowToGroup(d, cw, group);

			gw->inSelection = FALSE;
			damageScreen(cw->screen);
		}

		// exit selection
		free(gd->tmpSel.windows);
		gd->tmpSel.windows = NULL;
		gd->tmpSel.nWins = 0;
	}

	return FALSE;
}

/*
 * groupUnGroupWindows
 *
 */
Bool
groupUnGroupWindows(CompDisplay * d, CompAction * action,
		    CompActionState state, CompOption * option,
		    int nOption)
{
	CompWindow *cw = findWindowAtDisplay(d, d->activeWindow);

	if (!cw)
		return FALSE;

	GROUP_WINDOW(cw);

	if (gw->group) {
		groupDeleteGroup(d, gw->group);
	}

	return FALSE;
}

/*
 * groupRemoveWindow
 *
 */
Bool
groupRemoveWindow(CompDisplay * d, CompAction * action,
		  CompActionState state, CompOption * option, int nOption)
{
	CompWindow *cw = findWindowAtDisplay(d, d->activeWindow);
	if (!cw)
		return FALSE;
	GROUP_WINDOW(cw);

	if (gw->group)
		groupDeleteGroupWindow(d, cw);

	return FALSE;
}

/*
 * groupCloseWindows
 *
 */
Bool
groupCloseWindows(CompDisplay * d, CompAction * action,
		  CompActionState state, CompOption * option, int nOption)
{
	CompWindow *w = findWindowAtDisplay(d, d->activeWindow);
	if (!w)
		return FALSE;
	GROUP_WINDOW(w);

	if (gw->group) {

		int nWins = gw->group->nWins;
		int i;
		for (i = 0; i < nWins; i++) {
			CompWindow *cw = gw->group->windows[i];
			closeWindow(cw, getCurrentTimeFromDisplay(d));
		}
	}

	return FALSE;
}

/*
 * groupMinimizeWindows
 *
 */
Bool
groupMinimizeWindows(CompDisplay * d, CompAction * action,
		     CompActionState state, CompOption * option,
		     int nOption)
{
	CompWindow *w = findWindowAtDisplay(d, d->activeWindow);
	if (!w)
		return FALSE;
	GROUP_WINDOW(w);

	if (gw->group) {

		int nWins = gw->group->nWins;
		int i;
		for (i = 0; i < nWins; i++) {
			CompWindow *cw = gw->group->windows[i];
			minimizeWindow(cw);
		}
	}

	return FALSE;
}

/*
 * groupChangeColor
 *
 */
Bool
groupChangeColor(CompDisplay * d, CompAction * action,
		     CompActionState state, CompOption * option,
		     int nOption)
{
	CompWindow *w = findWindowAtDisplay(d, d->activeWindow);
	if (!w)
		return FALSE;
	GROUP_WINDOW(w);

	if (gw->group) {
		srand(time(NULL));
		gw->group->color[0] = rand() % 0xFFFF;
		gw->group->color[1] = rand() % 0xFFFF;
		gw->group->color[2] = rand() % 0xFFFF;

		groupRenderTopTabHighlight(w->screen, gw->group);

		damageScreen(w->screen);
	}

	return FALSE;
}

/*
 * groupSetIgnore
 *
 */
Bool
groupSetIgnore(CompDisplay * d, CompAction * action, CompActionState state,
	       CompOption * option, int nOption)
{
	GROUP_DISPLAY(d);

	gd->ignoreMode = TRUE;

	if (state & CompActionStateInitKey)
		action->state |= CompActionStateTermKey;

	return FALSE;
}

/*
 * groupUnsetIgnore
 *
 */
Bool
groupUnsetIgnore(CompDisplay * d, CompAction * action,
		 CompActionState state, CompOption * option, int nOption)
{
	GROUP_DISPLAY(d);

	gd->ignoreMode = FALSE;

	action->state &= ~CompActionStateTermKey;

	return FALSE;
}

/*
 * groupHandleButtonPressEvent
 *
 */
static void
groupHandleButtonPressEvent(CompDisplay *d, XEvent *event)
{
	int xRoot = event->xbutton.x_root;
	int yRoot = event->xbutton.y_root;
	int button = event->xbutton.button;

	GROUP_DISPLAY(d);

	if (button == 1) {
		GroupSelection *group;
			
		for (group = gd->groups; group; group = group->next) {
			if (group->inputPrevention.id == event->xbutton.window) {
				GroupTabBarSlot *slot;
					
				for (slot = group->tabBar->slots; slot; slot = slot->next) {
					if (XPointInRegion (slot->region, xRoot, yRoot)) {
						CompScreen *s = TOP_TAB(group)->screen;
						GROUP_SCREEN(s);

						gs->draggedSlot = slot;
						gs->dragged = FALSE; // The slot isn't dragged yet...
							
						gs->prevX = xRoot;
						gs->prevY = yRoot;

						if (!gs->grabIndex && !otherScreenGrabExist(s, "group", 0)) 
							gs->grabIndex = pushScreenGrab(s, None, "group");
					}
				}
			}
		}
	} else if ((button == 4) || (button == 5)) {
		CompWindow *w = findWindowAtDisplay(d, d->activeWindow);
		CompWindow *topTab = w;
			
		if(!w)
			return;

		GROUP_WINDOW(w);
		GROUP_SCREEN(w->screen);
	
		if(!gw->slot || !gw->group || !gw->group->tabBar)
			return;

		if(gw->group->tabBar->state != PaintOn && gw->group->tabBar->state != PaintPermanentOn)
			return;

		if(!XPointInRegion(gw->group->tabBar->region, xRoot, yRoot))
			return;
			
		if(gw->group->nextTopTab)
			topTab = NEXT_TOP_TAB(gw->group);
		else if(gw->group->topTab)
			topTab = TOP_TAB(gw->group); // If there are no tabbing animations, topTab is never NULL.
			
		gw = GET_GROUP_WINDOW(topTab, gs);

		if (button == 4) {
			if(gw->slot->prev)
				groupChangeTab(gw->slot->prev, RotateLeft);
			else
				groupChangeTab(gw->group->tabBar->revSlots, RotateLeft);
		} else {
			if(gw->slot->next)
				groupChangeTab(gw->slot->next, RotateRight);
			else
				groupChangeTab(gw->group->tabBar->slots, RotateRight);
		}
	}
}

/*
 * groupHandleButtonReleaseEvent
 *
 */
static void
groupHandleButtonReleaseEvent(CompDisplay *d, XEvent *event)
{
	GroupSelection *group;

	if (event->xbutton.button != 1)
		return;

	CompScreen *s = findScreenAtDisplay(d, event->xbutton.root);
	if (!s) 
		return;
	
	GROUP_SCREEN(s);
	GROUP_DISPLAY(d);

	if(!gs->draggedSlot)
		return;
			
	if(!gs->dragged) {
		groupChangeTab(gs->draggedSlot, RotateUncertain);
		gs->draggedSlot = NULL;
				
		if (gs->grabIndex) {
			removeScreenGrab(s, gs->grabIndex, NULL);
			gs->grabIndex = 0;
		}
		return;
	}
			
	GROUP_WINDOW(gs->draggedSlot->window);
	
	Bool inserted = FALSE;
	for(group = gd->groups; group; group = group->next) {
		Bool inTabBar;
		
		if (!group->tabBar)
			continue;
		
		Region buf = XCreateRegion();
		XIntersectRegion(gs->draggedSlot->region, group->tabBar->region, buf);
		
		inTabBar = !XEmptyRegion(buf);
		
		XDestroyRegion(buf);
		
		if(!inTabBar)
			continue;

		GroupTabBarSlot *slot;
		
		for(slot = group->tabBar->slots; slot; slot = slot->next) {
			if(slot == gs->draggedSlot)
				continue;
			
			Region slotRegion = XCreateRegion();
			XRectangle rect;
			Bool inSlot;
			
			if(slot->prev && slot->prev != gs->draggedSlot)
				rect.x = slot->prev->region->extents.x2;
			else if(slot->prev && slot->prev == gs->draggedSlot && gs->draggedSlot->prev)
				rect.x = gs->draggedSlot->prev->region->extents.x2;
			else
				rect.x = group->tabBar->region->extents.x1;
			
			rect.y = slot->region->extents.y1;
			
			if(slot->next && slot->next != gs->draggedSlot)
				rect.width = slot->next->region->extents.x1 - rect.x;
			else if(slot->next && slot->next == gs->draggedSlot && gs->draggedSlot->next)
				rect.width = gs->draggedSlot->next->region->extents.x1 - rect.x;
			else
				rect.width = group->tabBar->region->extents.x2;
			
			rect.height = slot->region->extents.y2 - slot->region->extents.y1;
			
			XUnionRectWithRegion(&rect, slotRegion, slotRegion);
			
			Region buf = XCreateRegion();
			XIntersectRegion(gs->draggedSlot->region, slotRegion, buf);
			
			inSlot = !XEmptyRegion(buf);
			
			XDestroyRegion(buf);
			XDestroyRegion(slotRegion);

			if (!inSlot)
				continue;
			
			GroupTabBarSlot *tmpDraggedSlot = gs->draggedSlot;
					
			if(group != gw->group) {
				GroupSelection *oldGroup = gw->group;
				// if the dragged window is not the top tab, move it onscreen
				if (!IS_TOP_TAB(gs->draggedSlot->window, oldGroup) && oldGroup->topTab) {
					CompWindow *tw = oldGroup->topTab->window;

					oldGroup->oldTopTabCenterX = WIN_X(tw) + WIN_WIDTH(tw) / 2;
					oldGroup->oldTopTabCenterY = WIN_Y(tw) + WIN_HEIGHT(tw) / 2;

					groupMoveTabOffscreen(gs->draggedSlot->window, oldGroup, FALSE);
				}

				// Change the group.
				groupAddWindowToGroup(s->display, gs->draggedSlot->window, group);
				groupTabSetVisibility(oldGroup, FALSE, PERMANENT);
			} else {
				groupUnhookTabBarSlot(group->tabBar, gs->draggedSlot, TRUE);
				groupTabSetVisibility(group, FALSE, PERMANENT);
			}

			// for re-calculating the tab-bar including the dragged window
			gs->draggedSlot = NULL;
			gs->dragged = FALSE;
			inserted = TRUE;
				
			if((tmpDraggedSlot->region->extents.x1 + tmpDraggedSlot->region->extents.x2) / 2 > 
				    (slot->region->extents.x1 + slot->region->extents.x2) / 2)
				groupInsertTabBarSlotAfter(group->tabBar, tmpDraggedSlot, slot);
			else
				groupInsertTabBarSlotBefore(group->tabBar, tmpDraggedSlot, slot);

			break;
		}

		if (inserted)
			break;
	}

	if (!inserted) {
		GROUP_WINDOW(gs->draggedSlot->window);
		groupTabSetVisibility(gw->group, FALSE, PERMANENT);
				
		gs->draggedSlot = NULL;
		gs->dragged = FALSE;

		if (gw->group->topTab)
			groupRecalcTabBarPos(gw->group,
				(gw->group->tabBar->region->extents.x1 + gw->group->tabBar->region->extents.x2) / 2,
				gw->group->tabBar->region->extents.x1, gw->group->tabBar->region->extents.x2);

		// to remove the painted slot
		damageScreen(s);
	}

	if (gs->grabIndex) {
		removeScreenGrab(s, gs->grabIndex, NULL);
		gs->grabIndex = 0;
	}

	if (gs->dragHoverTimeoutHandle) {
		compRemoveTimeout(gs->dragHoverTimeoutHandle);
		gs->dragHoverTimeoutHandle = 0;
	}
}

/*
 * groupHandleMotionEvent
 *
 */

/* the radius to determine if it was a click or a drag */
#define RADIUS 5

static void 
groupHandleMotionEvent(CompScreen * s, int xRoot, int yRoot)
{
	GROUP_SCREEN(s);

	if (gs->grabIndex) {
		if (gs->draggedSlot) {
			int dx, dy;
			REGION reg;
			Region draggedRegion = gs->draggedSlot->region;
								
			reg.rects = &reg.extents;
			reg.numRects = 1;

			dx = xRoot - gs->prevX;
			dy = yRoot - gs->prevY;
					
			if(gs->dragged || abs(dx) > RADIUS || abs(dy) > RADIUS) {
				gs->prevX = xRoot;
				gs->prevY = yRoot;
						
				if(!gs->dragged) {
					GROUP_WINDOW(gs->draggedSlot->window);
						
					gs->dragged = TRUE;
					groupTabSetVisibility(gw->group, TRUE, PERMANENT);
					
					groupRecalcTabBarPos(gw->group, (gw->group->tabBar->region->extents.x1 + gw->group->tabBar->region->extents.x2) / 2,
			    						 gw->group->tabBar->region->extents.x1, gw->group->tabBar->region->extents.x2);
				}
						
				// (border = 5) + (buffer = 2) == (damage buffer = 7)
				//TODO: respect border option.
				reg.extents.x1 = draggedRegion->extents.x1 - 7;
				reg.extents.y1 = draggedRegion->extents.y1 - 7;
				reg.extents.x2 = draggedRegion->extents.x2 + 7;
				reg.extents.y2 = draggedRegion->extents.y2 + 7;
				damageScreenRegion(s, &reg);
									
				XOffsetRegion (gs->draggedSlot->region, dx, dy);
				gs->draggedSlot->springX = (gs->draggedSlot->region->extents.x1 + gs->draggedSlot->region->extents.x2) / 2;

				reg.extents.x1 = draggedRegion->extents.x1 - 7;
				reg.extents.y1 = draggedRegion->extents.y1 - 7;
				reg.extents.x2 = draggedRegion->extents.x2 + 7;
				reg.extents.y2 = draggedRegion->extents.y2 + 7;
				damageScreenRegion(s, &reg);
			}
		} else 
			groupDamageSelectionRect(s, xRoot, yRoot);
	}
}

/*
 * groupHandleEvent
 *
 */
void groupHandleEvent(CompDisplay * d, XEvent * event)
{
	GROUP_DISPLAY(d);

	switch (event->type) {
	case MotionNotify:
		{
			CompScreen *s = findScreenAtDisplay(d, event->xmotion.root);
			if (s)
				groupHandleMotionEvent(s, d->pointerX, d->pointerY);
			
			break;
		}

	case ButtonPress:
		groupHandleButtonPressEvent(d, event);
		break;

	case ButtonRelease:
		groupHandleButtonReleaseEvent(d, event);
		break;

	case UnmapNotify:
		{
			CompWindow *w = findWindowAtDisplay(d, event->xunmap.window);
			if (!w)
				break;
			GROUP_WINDOW(w);

			if (w->pendingUnmaps) {
				if (w->shaded)
					gw->windowState = WindowShaded;
				else if (!w->invisible)
					gw->windowState = WindowMinimized;
			}

			if (gw->group) {
				if (gw->group->tabBar && IS_TOP_TAB(w, gw->group)) {
					/* on unmap of the top tab, hide the tab bar and the 
					   input prevention window */
					groupTabSetVisibility(gw->group, FALSE, PERMANENT);
				}
				if (!w->pendingUnmaps) {
					//close event
					groupDeleteGroupWindow(d, w);
					damageScreen(w->screen);
				}
			}
			break;
		}

	default:
		break;
	}

	UNWRAP(gd, d, handleEvent);
	(*d->handleEvent) (d, event);
	WRAP(gd, d, handleEvent, groupHandleEvent);

	switch (event->type) {
	case PropertyNotify:
		// focus event
		if (event->xproperty.atom == d->winActiveAtom) {
			CompWindow *w = (CompWindow *) findWindowAtDisplay(d, d->activeWindow);
			if (!w)
				break;

			GROUP_WINDOW(w);
			GROUP_SCREEN(w->screen);

			if (gw->group && !gw->group->tabBar &&
			    !screenGrabExist(w->screen, "scale", 0) &&
			    gs->opt[GROUP_SCREEN_OPTION_RAISE].value.b) {
				groupRaiseWindows(w, gw->group);
			} else if(gw->group && gw->group->tabBar && gw->group->topTab) {
				focusWindow(TOP_TAB(gw->group));
				moveInputFocusToWindow(TOP_TAB(gw->group));
			}

		}
		break;

	case ConfigureNotify:
		{
			CompWindow *w = findWindowAtDisplay(d, event->xconfigure.window);
			if (!w)
				break;

			GROUP_WINDOW(w);
			if (gw->group && (gw->group->grabWindow != None)) {
				if (!gw->group->grabMask) 
					gw->group->grabWindow = None;
			}
			break;
		}

	case EnterNotify:
		{
			CompWindow *w = findWindowAtDisplay(d, event->xcrossing.window);
			if (!w)
				break;

			GROUP_WINDOW(w);
			if (gw->group) {
				GROUP_SCREEN(w->screen);

				if (gs->draggedSlot && gs->dragged && IS_TOP_TAB(w, gw->group)) {
					int hoverTime = 
						gs->opt[GROUP_SCREEN_OPTION_TAB_DRAG_HOVER_TIME].value.f * 1000;
					if (gs->dragHoverTimeoutHandle)
						compRemoveTimeout(gs->dragHoverTimeoutHandle);

					if (hoverTime > 0)
						gs->dragHoverTimeoutHandle = 
							compAddTimeout(hoverTime, groupDragHoverTimeout, w);
				}
			}
		}

	default:
		break;
	}
}

/*
 * groupGetOutputExtentsForWindow
 *
 */
void
groupGetOutputExtentsForWindow (CompWindow * w, CompWindowExtents * output)
{
	GROUP_SCREEN (w->screen);
	GROUP_WINDOW (w);

	UNWRAP (gs, w->screen, getOutputExtentsForWindow);
	(*w->screen->getOutputExtentsForWindow) (w, output);
	WRAP (gs, w->screen, getOutputExtentsForWindow,
		groupGetOutputExtentsForWindow);

	if (gw->group)
	{
		int glowSize = gs->opt[GROUP_SCREEN_OPTION_GLOW_SIZE].value.i;

		if (glowSize > output->left)
			output->left = glowSize;
		if (glowSize > output->right)
			output->right = glowSize;
		if (glowSize > output->top)
			output->top = glowSize;
		if (glowSize > output->bottom)
			output->bottom = glowSize;
	}
}

/*
 * groupWindowResizeNotify
 *
 */
void groupWindowResizeNotify(CompWindow * w, int dx, int dy, int dwidth, int dheight)
{
	GROUP_SCREEN(w->screen);
	GROUP_DISPLAY(w->screen->display);
	GROUP_WINDOW(w);

	UNWRAP(gs, w->screen, windowResizeNotify);
	(*w->screen->windowResizeNotify) (w, dx, dy, dwidth, dheight);
	WRAP(gs, w->screen, windowResizeNotify, groupWindowResizeNotify);

	if (gw->glowQuads)
		groupComputeGlowQuads (w, &gs->glowTexture.matrix);

	if (!dx && !dy && !dwidth && !dheight)
		return;

	if (gw->group && gw->group->tabBar && 
	    (gw->group->tabBar->state != PaintOff) && IS_TOP_TAB(w, gw->group)) {
		groupRecalcTabBarPos(gw->group, w->screen->display->pointerX,
			WIN_X(w), WIN_X(w) + WIN_WIDTH(w));
	}

	if (gw->group && (gw->group->grabWindow == w->id) && 
	    !gd->ignoreMode && gs->opt[GROUP_SCREEN_OPTION_RESIZE].value.b) {
		int i;
		for (i = 0; i < gw->group->nWins; i++) {
			CompWindow *cw =  gw->group->windows[i];
			if (!cw)
				continue;

			if (cw->id != w->id) {
				int nx = 0;
				int ny = 0;
	
				if (gs->opt[GROUP_SCREEN_OPTION_RELATIVE_DISTANCE].value.b) {
					int distX = WIN_X(cw) - (WIN_X(w) - dx);
					int distY = WIN_Y(cw) - (WIN_Y(w) - dy);
					int ndistX = distX * ((float) WIN_WIDTH(w) / (float) (WIN_WIDTH(w) - dwidth));
					int ndistY = distY * ((float) w->serverHeight / (float) (WIN_HEIGHT(w) - dheight));
					nx = WIN_X(w) + ndistX;
					ny = WIN_Y(w) + ndistY;
				} else {
					nx = WIN_X(cw) + dx;
					ny = WIN_Y(cw) + dy;
				}
	
				int nwidth = WIN_WIDTH(cw) + dwidth;
				int nheight = WIN_HEIGHT(cw) + dheight;
				constrainNewWindowSize(cw, nwidth, nheight,
						&nwidth, &nheight);

				XWindowChanges xwc;

				if(gw->group->tabBar)
				{
					xwc.x = nx;
					xwc.y = ny;
				}
				
				else
				{
					xwc.x = cw->serverX;
					xwc.y = cw->serverY;
				}
				
				xwc.width = nwidth;
				xwc.height = nheight;

				configureXWindow(cw, CWX | CWY | CWWidth | CWHeight, &xwc);
			}
		}
	}
}

/*
 * groupWindowMoveNotify
 *
 */
void
groupWindowMoveNotify(CompWindow * w, int dx, int dy, Bool immediate)
{
	GROUP_SCREEN(w->screen);
	GROUP_DISPLAY(w->screen->display);
	GROUP_WINDOW(w);

	UNWRAP(gs, w->screen, windowMoveNotify);
	(*w->screen->windowMoveNotify) (w, dx, dy, immediate);
	WRAP(gs, w->screen, windowMoveNotify, groupWindowMoveNotify);

	if (gw->glowQuads)
		groupComputeGlowQuads (w, &gs->glowTexture.matrix);

	if(!gw->group || gs->queued)
		return;
	
	Bool viewportChange = screenGrabExist (w->screen, "rotate", 0) &&
		((dx && !(dx % w->screen->width)) || (dy && !(dy % w->screen->height)));

	if (viewportChange && (gw->animateState & IS_ANIMATED)) {
		gw->destination.x += dx;
		gw->destination.y += dy;
	}

	if (gw->group->tabBar && gw->group->topTab)
	{
		if (TOP_TAB(gw->group)->id != w->id) {
			if (!otherScreenGrabExist(w->screen, 0))
				groupEnqueueMoveNotify(w, -dx, -dy, TRUE);
			else
				groupEnqueueMoveNotify(w, -dx, -dy, FALSE);
		} else {
			GroupTabBarSlot *slot;
			GroupTabBar *bar = gw->group->tabBar;
			if(gs->opt[GROUP_SCREEN_OPTION_SPRING_MODEL_ON_MOVE].value.b)
				XOffsetRegion (bar->region, 0, dy);
			else
				XOffsetRegion (bar->region, dx, dy);
			bar->rightSpringX += dx;
			bar->leftSpringX += dx;
			
			for(slot = bar->slots; slot; slot = slot->next)
			{
				if(gs->opt[GROUP_SCREEN_OPTION_SPRING_MODEL_ON_MOVE].value.b)
					XOffsetRegion (slot->region, 0, dy);
				else
					XOffsetRegion (slot->region, dx, dy);
				slot->springX += dx;
			}
			
			groupUpdateInputPreventionWindow(gw->group);
		}

		return;
	}

	if (gw->group->doTabbing || viewportChange || 
	    gd->ignoreMode || (gw->group->grabWindow != w->id) ||
	    !screenGrabExist (w->screen, "move", 0) ||
	    !gs->opt[GROUP_SCREEN_OPTION_MOVE].value.b)
		return;

	int i;

	for (i = 0; i < gw->group->nWins; i++) {
		CompWindow *cw = gw->group->windows[i];
		if (!cw)
			continue;

		if (cw->id != w->id) {
			GROUP_WINDOW(cw);
	
			gw->needsPosSync = TRUE;
			groupEnqueueMoveNotify (cw, dx, dy, FALSE);
		}
	}
}

void
groupWindowGrabNotify(CompWindow * w,
                       int x, int y, unsigned int state, unsigned int mask)
{
	GROUP_SCREEN(w->screen);
	GROUP_DISPLAY(w->screen->display);
	GROUP_WINDOW(w);

	if (gw->group && !gd->ignoreMode && !gs->queued) {
		int i;
		for (i = 0; i < gw->group->nWins; i++) {
			CompWindow *cw = gw->group->windows[i];
			if (!cw)
				continue;

			if (cw->id != w->id) 
				groupEnqueueGrabNotify (cw, x, y, state, mask);
		}

		gw->group->grabWindow = w->id;
		gw->group->grabMask = mask;
	}

	UNWRAP(gs, w->screen, windowGrabNotify);
	(*w->screen->windowGrabNotify) (w, x, y, state, mask);
	WRAP(gs, w->screen, windowGrabNotify, groupWindowGrabNotify);
}

void groupWindowUngrabNotify(CompWindow * w)
{
	GROUP_SCREEN(w->screen);
	GROUP_DISPLAY(w->screen->display);
	GROUP_WINDOW(w);

	if (gw->group && !gd->ignoreMode && !gs->queued) {
		int i;

		groupDequeueMoveNotifies(w->screen);

		for (i = 0; i < gw->group->nWins; i++) {
			CompWindow *cw = gw->group->windows[i];
			if (!cw)
				continue;

			if (cw->id != w->id) {
				GROUP_WINDOW(cw);

				if (gw->needsPosSync) {
					syncWindowPosition(cw);
					gw->needsPosSync = FALSE;
				}
				groupEnqueueUngrabNotify (cw);
			}
		}

		if (!(gw->group->grabMask & CompWindowGrabResizeMask))	
			gw->group->grabWindow = None;
		

		gw->group->grabMask = 0;
	}

	UNWRAP(gs, w->screen, windowUngrabNotify);
	(*w->screen->windowUngrabNotify) (w);
	WRAP(gs, w->screen, windowUngrabNotify, groupWindowUngrabNotify);
}


Bool groupDamageWindowRect(CompWindow * w, Bool initial, BoxPtr rect)
{

	GROUP_SCREEN(w->screen);
	Bool status;

	UNWRAP(gs,w->screen,damageWindowRect);
	status = (*w->screen->damageWindowRect) (w,initial,rect);
	WRAP(gs,w->screen,damageWindowRect,groupDamageWindowRect);

	if (initial) {
		GROUP_WINDOW(w);

		if (gs->opt[GROUP_SCREEN_OPTION_AUTOTAB].value.b && (w->type & gs->wMask)) {
			if (!gw->group && (gw->windowState == WindowNormal)) {
				groupAddWindowToGroup(w->screen->display, w, NULL);
				groupTabGroup(w);
			}
		}

		gw->windowState = WindowNormal;
	}
	return status;

}
